@@ -1,10 +1,12 @@
-//! Gameloop and [gameloop::GameControl] context, which is used to control the game globally
+//! Gameloop and [GameControl] context, which is used to control the game globally
 
+use crate::datapack::DataMultistore;
+use crate::datapack::source::SourceManager;
 use crate::input::Input;
 use crate::scene::{BaseScene, NullSpawnable, Scene, SceneConfig, SceneType};
-use crate::audio::{AlAudio, Audio, AudioError, NullAudio};
-use crate::renderer::{Renderer, RendererError, RendererNewInfo, ViewportMode, WindowScaleMode};
-#[cfg(feature = "imgui_base")] use crate::imgui_handler::ImGuiHandler;
+#[cfg(feature = "audio")] use crate::audio::{AlAudio, Audio, AudioError, NullAudio};
+use crate::renderer::{Renderer, RendererError, ViewportMode, WindowScaleMode};
+#[cfg(feature = "imgui")] use crate::imgui_handler::ImGuiHandler;
 #[cfg(feature = "default_logger")] use crate::doglog;
 
 use sdl2::event::Event;
@@ -15,10 +17,7 @@ use sdl2::messagebox;
 #[cfg(feature = "graphical_panic")] use std::panic::PanicInfo;
 #[cfg(feature = "graphical_panic")] use backtrace::Backtrace;
 
-use keeshond_datapack::{DataId, DataMultistore, DataObject};
-use keeshond_datapack::source::SourceManager;
-
-#[cfg(feature = "imgui_base")] use imgui;
+#[cfg(feature = "imgui")] use imgui;
 
 use std::time::{Duration, Instant};
 use std::rc::Rc;
@@ -28,12 +27,12 @@ use std::any::TypeId;
 use downcast_rs::Downcast;
 use std::sync::{Arc, Mutex, MutexGuard};
 use crate::doglog::ConsoleHistory;
+use crate::renderer::draw_control_private::RendererNewInfo;
 
 const MILLISECONDS_PER_NANOSECOND : f64 = 0.000001;
 const BUSY_WAIT_MARGIN : u64 = 2; // For smooth results on Windows. Linux can probably get away with 1.
 const AUTOMATIC_LERP_CAP_MULTIPLIER : f64 = 1.05;
 const MAX_POST_SLOWDOWN_CATCHUP : u32 = 10;
-const MAX_LOGIC_FRAME_OVERRUN : u32 = 8;
 const MAX_FRAME_JIFFY_SUBTRACT : u32 = 1000;
 const MAX_FPS_STAT_CATCHUP : u64 = 3;
 const LOW_ENERGY_TIMER_MIN : u32 = 60;
@@ -55,6 +54,7 @@ pub enum InitError
     Sdl2InitError(Sdl2Subsystem, String),
     #[fail(display = "{}", _0)]
     RendererInitError(RendererError),
+    #[cfg(feature = "audio")]
     #[fail(display = "{}", _0)]
     AudioInitError(AudioError),
     #[fail(display = "{}", _0)]
@@ -125,7 +125,7 @@ fn print_log_section(title : &str)
     info!("");
 }
 
-/// (ADVANCED) Displays a message popup in the GUI and prints the message to the console
+/// Displays a message popup in the GUI and prints the message to the console
 #[allow(unused_variables)]
 pub fn gui_cli_message(message : &str, title : &str)
 {
@@ -144,12 +144,10 @@ fn read_options() -> getopts::Matches
 {
     let args : Vec<String> = std::env::args().collect();
     let mut options = getopts::Options::new();
-    let backends = crate::renderer::available_backends();
     
     options.optflag("", "help", "Show this help");
     options.optflag("f", "fullscreen", "Run in fullscreen mode");
     options.optflag("w", "window", "Run in windowed mode");
-    options.optopt("b", "backends", &format!("Select the renderer backends to use, separated by spaces and ordered by preference. Default list: \"{}\"", backends), "BACKEND_LIST");
     options.optflag("", "interpolate", "When an alternate refresh rate is detected, unlock display framerate from logic and enable frame interpolation.");
     options.optflag("", "no-interpolate", "Force disable frame interpolation and lock display rate to logic rate. Use when monitor's refresh rate equals game's logic rate.");
     options.optflag("", "force-interpolate", "Force enable frame interpolation. Not recommended if monitor's refresh rate equals game's logic rate!");
@@ -227,8 +225,6 @@ fn apply_config(gameinfo : &mut GameInfo)
         },
         _ => ()
     };
-
-    gameinfo.backends = matches.opt_str("backends").unwrap_or(crate::renderer::available_backends());
     
     if matches.opt_present("kiosk")
     {
@@ -301,10 +297,6 @@ pub struct GameInfo
     ///  Also pauses the application when it is not focused unless there is interaction.
     ///  Useful for editors and other utilities.
     pub low_energy_app : bool,
-    /// A space-delimited list of rendering backends to try, in order of preference. If the first
-    ///  backend cannot be initialized, the second one is attempted, and etc. Recommend keeping
-    ///  this at the default unless you know what you're doing.
-    pub backends : String,
     /// Whether to print basic stats (number of entities and components, framerate, basic timing information)
     ///  to the log. Does nothing on release builds.
     pub log_stats : bool,
@@ -338,14 +330,13 @@ impl Default for GameInfo
             target_framerate : 60.0,
             frame_interpolation : FrameInterpolationMode::Off,
             frame_interpolation_cap : FrameInterpolationCap::Automatic,
-            backends : crate::renderer::available_backends(),
             log_stats : true,
             log_packages : Vec::new()
         }
     }
 }
 
-/// (ADVANCED) Provides basic diagnostic timekeeping facilities
+/// (Advanced) Provides basic diagnostic timekeeping facilities
 pub struct GameloopStopwatch
 {
     jiffies : Instant,
@@ -455,6 +446,8 @@ impl GameloopStopwatch
     }
 }
 
+/// Trait for a struct that stores data global to the game - i.e. presists between scene switches,
+///  and is accessible using [GameControl::singleton_mut()].
 pub trait GameSingleton : Downcast
 {
 
@@ -462,7 +455,7 @@ pub trait GameSingleton : Downcast
 
 impl_downcast!(GameSingleton);
 
-pub struct GameControlOptions<'a>
+pub(crate) struct GameControlOptions<'a>
 {
     console_history: Arc<Mutex<ConsoleHistory>>,
     sdl_context : Option<&'a sdl2::Sdl>
@@ -470,6 +463,7 @@ pub struct GameControlOptions<'a>
 
 impl<'a> GameControlOptions<'a>
 {
+    /// Returns a configuration for creating the [GameControl] in headless mode (i.e. no window, renderer, or audio)
     pub fn headless() -> GameControlOptions<'a>
     {
         GameControlOptions
@@ -480,9 +474,9 @@ impl<'a> GameControlOptions<'a>
     }
 }
 
-/// Main context. You don't normally need to create this yourself as the [Gameloop]
-///  has its own instance. It can be useful to create this directly if you need to run
-///  automated tests.
+/// Main context, for controlling the game while it is running. You don't normally need to create
+///  this yourself as [Gameloop::new()] does this for you, but it can be useful to create this
+///  directly if you need to run automated tests.
 pub struct GameControl
 {
     target_framerate : f64,
@@ -493,6 +487,7 @@ pub struct GameControl
     auto_unload : bool,
     input : Input,
     renderer : Renderer,
+    #[cfg(feature = "audio")]
     audio : Box<dyn Audio>,
     #[allow(dead_code)]
     controller_subsystem : Option<sdl2::GameControllerSubsystem>,
@@ -506,34 +501,37 @@ pub struct GameControl
     stats : GameStats,
     focused : bool,
     done : bool,
-    #[cfg(feature = "imgui_base")]
+    #[cfg(feature = "imgui")]
     imgui_handler : Option<ImGuiHandler>
 }
 
 impl GameControl
 {
-    /// (ADVANCED) Creates the main game context
-    pub fn new(gameinfo : &GameInfo, options : GameControlOptions) -> Result<GameControl, InitError>
+    pub(crate) fn new(gameinfo : &GameInfo, options : GameControlOptions) -> Result<GameControl, InitError>
     {   
         let source_manager = Rc::new(RefCell::new(SourceManager::new()));
         
         let mut resources = DataMultistore::new(source_manager.clone());
         
         let mut renderer_new_info = None;
+        #[cfg(feature = "audio")]
         let audio : Box<dyn Audio>;
         let controller_subsystem;
         let mut input;
 
         if let Some(sdl) = options.sdl_context
         {
-            let video_subsystem = try_or_else!(sdl.video(),
-                |error| Err(InitError::Sdl2InitError(Sdl2Subsystem::Video, error)));
+            let video_subsystem = sdl.video().map_err(
+                |error| InitError::Sdl2InitError(Sdl2Subsystem::Video, error))?;
 
-            controller_subsystem = Some(try_or_else!(sdl.game_controller(),
-                |error| Err(InitError::Sdl2InitError(Sdl2Subsystem::GameController, error))));
+            controller_subsystem = Some(sdl.game_controller().map_err(
+                |error| InitError::Sdl2InitError(Sdl2Subsystem::GameController, error))?);
 
-            audio = Box::new(try_or_else!(AlAudio::new(&mut resources),
-                |error| Err(InitError::AudioInitError(error))));
+            #[cfg(feature = "audio")]
+            {
+                audio = Box::new(AlAudio::new(&mut resources).map_err(
+                    |error| InitError::AudioInitError(error))?);
+            }
 
             input = Input::new(Some(sdl.mouse()));
             input.set_cursor_visible(gameinfo.cursor_visible);
@@ -557,28 +555,30 @@ impl GameControl
                 fullscreen : gameinfo.fullscreen,
                 viewport_mode : gameinfo.viewport_mode,
                 video_subsystem,
-                backends : gameinfo.backends.clone(),
                 resources : &mut resources
             });
         }
         else
         {
-            audio = Box::new(NullAudio::new());
+            #[cfg(feature = "audio")]
+            {
+                audio = Box::new(NullAudio::new());
+            }
             input = Input::new(None);
             controller_subsystem = None;
         }
 
-        #[cfg(feature = "imgui_base")]
-        let mut imgui_handler = try_or_else!(ImGuiHandler::new(),
-            |error| Err(InitError::ImguiInitError(error)));
+        #[cfg(feature = "imgui")]
+        let mut imgui_handler = ImGuiHandler::new().map_err(
+            |error| InitError::ImguiInitError(error))?;
         
         #[allow(unused_mut)] // mut used only by imgui feature
-        let mut renderer = try_or_else!(Renderer::new(renderer_new_info),
-            |error| Err(InitError::RendererInitError(error)));
+        let mut renderer = Renderer::new(renderer_new_info).map_err(
+            |error| InitError::RendererInitError(error))?;
         
-        #[cfg(feature = "imgui_base")]
-        try_or_else!(renderer.control().init_imgui_renderer(imgui_handler.imgui_mut()),
-            |error| Err(InitError::RendererInitError(error)));
+        #[cfg(feature = "imgui")]
+        renderer.control().init_imgui_renderer(imgui_handler.imgui_mut()).map_err(
+            |error| InitError::RendererInitError(error))?;
 
         input.update_mouse_metrics(&renderer);
         
@@ -592,6 +592,7 @@ impl GameControl
             auto_unload : gameinfo.auto_unload,
             input,
             renderer,
+            #[cfg(feature = "audio")]
             audio,
             controller_subsystem,
             controllers : HashMap::new(),
@@ -604,46 +605,163 @@ impl GameControl
             focused : true,
             done : false,
             singletons : HashMap::new(),
-            #[cfg(feature = "imgui_base")]
+            #[cfg(feature = "imgui")]
             imgui_handler : Some(imgui_handler),
         })
     }
-    
+
+    /// Creates a [GameControl] in a headless configuration, allowing you to run automated tests,
+    ///  validation, or other processes without needing a gameloop, window, renderer, or audio.
+    pub fn new_headless(gameinfo : &GameInfo) -> Result<GameControl, InitError>
+    {
+        GameControl::new(gameinfo, GameControlOptions::headless())
+    }
+
+    /// Retrieves the [Input] object for reading player input such as keyboard, mouse, and gamepad.
+    pub fn input(&self) -> &Input
+    {
+        &self.input
+    }
+
+    /// Retrieves the [Input] object mutably for changing input bindings and deadzones
+    pub fn input_mut(&mut self) -> &mut Input
+    {
+        &mut self.input
+    }
+
+    /// Retrieves the [Renderer] object to read various parameters for the renderer window
+    pub fn renderer(&self) -> &Renderer
+    {
+        &self.renderer
+    }
+
+    /// Retrieves the [Renderer] object mutably to change various parameters for the renderer window
+    pub fn renderer_mut(&mut self) -> &mut Renderer
+    {
+        &mut self.renderer
+    }
+
+    /// Retrieves the [Audio] object for reading information about audio playback
+    #[cfg(feature = "audio")]
+    pub fn audio(&self) -> &Box<dyn Audio>
+    {
+        &self.audio
+    }
+
+    /// Retrieves the [Audio] object mutably to control audio playback
+    #[cfg(feature = "audio")]
+    pub fn audio_mut(&mut self) -> &mut Box<dyn Audio>
+    {
+        &mut self.audio
+    }
+
+    #[cfg(feature = "audio")]
+    fn sync_audio_store(&mut self)
+    {
+        self.audio.sync_sound_store(&mut self.resources.store_mut())
+            .expect("Audio backend store sync failed");
+    }
+
+    /// Retrieves the [SourceManager] for registering and querying data [Sources](crate::datapack::source::Source)
+    pub fn source_manager(&self) -> Rc<RefCell<SourceManager>>
+    {
+        self.source_manager.clone()
+    }
+
+    /// Retrieves the [DataMultistore] for accessing data resources
+    pub fn res(&self) -> &DataMultistore
+    {
+        &self.resources
+    }
+
+    #[allow(dead_code)]
+    pub(crate) fn res_mut(&mut self) -> &mut DataMultistore
+    {
+        &mut self.resources
+    }
+
+    /// Retrieves the [GameSingleton] of the current type mutably. If it has not been created yet,
+    ///  it will be [Default]-constructed.
+    pub fn singleton_mut<T : 'static + GameSingleton + Default>(&mut self) -> &mut T
+    {
+        let type_id = TypeId::of::<T>();
+
+        if !self.singletons.contains_key(&type_id)
+        {
+            self.singletons.insert(type_id.clone(), Box::new(T::default()));
+        }
+
+        self.singletons.get_mut(&type_id).unwrap().downcast_mut::<T>().expect("Downcast failed!")
+    }
+
+    /// Retrieves the imgui context
+    #[cfg(feature = "imgui")]
+    pub fn imgui(&self) -> Option<&imgui::Context>
+    {
+        if let Some(imgui_handler) = &self.imgui_handler
+        {
+            return Some(imgui_handler.imgui());
+        }
+
+        None
+    }
+
+    /// Retrieves the imgui context mutably
+    #[cfg(feature = "imgui")]
+    pub fn imgui_mut(&mut self) -> Option<&mut imgui::Context>
+    {
+        if let Some(imgui_handler) = &mut self.imgui_handler
+        {
+            return Some(imgui_handler.imgui_mut());
+        }
+
+        None
+    }
+
+    /// Returns the framerate that game logic is synchronized to, in frames per second
     pub fn target_framerate(&self) -> f64
     {
         self.target_framerate
     }
-    
+
+    /// Sets the framerate that game logic is synchronized to, in frames per second
     pub fn set_target_framerate(&mut self, new_framerate : f64)
     {
         self.target_framerate = new_framerate.clamp(5.0, 1000.0);
     }
-    
+
+    /// Returns the mode to use for interpolating frames in the renderer.
     pub fn frame_interpolation(&self) -> FrameInterpolationMode
     {
         self.frame_interpolation
     }
-    
+
+    /// Sets the mode to use for interpolating frames in the renderer.
     pub fn set_frame_interpolation(&mut self, mode : FrameInterpolationMode)
     {
         self.frame_interpolation = mode;
     }
-    
+
+    /// Returns the framerate cap mode for when frame interpolation is enabled.
     pub fn frame_interpolation_cap(&self) -> FrameInterpolationCap
     {
         self.interpolation_cap
     }
-    
+
+    /// Sets the framerate cap mode for when frame interpolation is enabled.
     pub fn set_frame_interpolation_cap(&mut self, mut cap : FrameInterpolationCap)
     {
         if let FrameInterpolationCap::Manual(value) = cap
         {
             cap = FrameInterpolationCap::Manual(value.clamp(5.0, 1000.0));
         }
-        
+
         self.interpolation_cap = cap;
     }
 
+    /// Returns true if frame interpolation is currently active. For the automatic mode, returns
+    ///  true if the refresh rate of the window's current monitor does not match with the target
+    ///  logic framerate.
     pub fn frame_interpolation_active(&self) -> bool
     {
         match self.frame_interpolation
@@ -663,113 +781,23 @@ impl GameControl
             FrameInterpolationMode::Force => { return true; }
         }
     }
-    
-    pub fn input(&self) -> &Input
-    {
-        &self.input
-    }
-    
-    pub fn input_mut(&mut self) -> &mut Input
-    {
-        &mut self.input
-    }
-    
-    pub fn renderer(&self) -> &Renderer
-    {
-        &self.renderer
-    }
-    
-    pub fn renderer_mut(&mut self) -> &mut Renderer
-    {
-        &mut self.renderer
-    }
-    
-    pub fn audio(&self) -> &Box<dyn Audio>
-    {
-        &self.audio
-    }
-    
-    pub fn audio_mut(&mut self) -> &mut Box<dyn Audio>
-    {
-        &mut self.audio
-    }
-    
-    fn sync_audio_store(&mut self)
-    {
-        self.audio.sync_sound_store(&mut self.resources.store_mut())
-            .expect("Audio backend store sync failed");
-    }
-    
-    pub fn source_manager(&self) -> Rc<RefCell<SourceManager>>
-    {
-        self.source_manager.clone()
-    }
-    
-    pub fn res(&self) -> &DataMultistore
-    {
-        &self.resources
-    }
-
-    #[allow(dead_code)]
-    pub(crate) fn res_mut(&mut self) -> &mut DataMultistore
-    {
-        &mut self.resources
-    }
-
-    /// Returns a [DataId] for the data object with the given package and name. Avoid calling this repeatedly,
-    ///  as there is a performance penalty to doing so. Please store the [DataId] where it can be used later.
-    ///  Panics if the load failed.
-    pub fn lookup_res<D : DataObject + 'static>(&mut self, package_name : &str, pathname : &str) -> DataId<D>
-    {
-        let mut store = self.resources.store_mut::<D>();
-
-        store.get_id_mut(package_name, pathname).expect(&format!("Failed to load {}/{}", package_name, pathname))
-    }
-
-    pub fn singleton_mut<T : 'static + GameSingleton + Default>(&mut self) -> &mut T
-    {
-        let type_id = TypeId::of::<T>();
-
-        if !self.singletons.contains_key(&type_id)
-        {
-            self.singletons.insert(type_id.clone(), Box::new(T::default()));
-        }
-
-        self.singletons.get_mut(&type_id).unwrap().downcast_mut::<T>().expect("Downcast failed!")
-    }
-
-    #[cfg(feature = "imgui_base")]
-    pub fn imgui(&self) -> Option<&imgui::Context>
-    {
-        if let Some(imgui_handler) = &self.imgui_handler
-        {
-            return Some(imgui_handler.imgui());
-        }
-
-        None
-    }
-
-    #[cfg(feature = "imgui_base")]
-    pub fn imgui_mut(&mut self) -> Option<&mut imgui::Context>
-    {
-        if let Some(imgui_handler) = &mut self.imgui_handler
-        {
-            return Some(imgui_handler.imgui_mut());
-        }
-
-        None
-    }
 
+    /// Retrieves the [ConsoleHistory] object, which is useful for implementing an in-game console window
     pub fn console_history(&self) -> MutexGuard<'_, ConsoleHistory>
     {
         self.console_history.lock().unwrap()
     }
 
+    /// Retrieves the [GameStats] object, which contains timing statistics
     pub fn stats(&self) -> &GameStats
     {
         &self.stats
     }
 
+    /// Switches the game to a new scene of the given [SceneType] after the current frame is done.
+    ///  Doing this will perform automatic unloading of unused data resources (ones without an active
+    ///  [crate::datapack::DataHandle] or [crate::datapack::PackageUseToken]) if it is configured to
+    ///  do so via the [GameInfo] passed in during creation of the [Gameloop] or [GameControl].
     pub fn goto_new_scene<T : SceneType + 'static>(&mut self)
     {
         let next_scene = Box::new(Scene::<T>::new());
@@ -777,7 +805,8 @@ impl GameControl
         self.next_scene = next_scene;
         self.next_scene_waiting = true;
     }
-    
+
+    /// Leaves the main gameloop. This usually means closing the game.
     pub fn quit(&mut self)
     {
         self.done = true;
@@ -794,7 +823,7 @@ struct GameloopTimeState
     cap_frametime : Duration,
     frame_start_jiffies : Duration,
     last_display : Instant,
-    #[cfg(feature = "imgui_base")]
+    #[cfg(feature = "imgui")]
     last_imgui : Instant,
     lag_this_frame : u32,
     lag_this_second : u32,
@@ -824,7 +853,7 @@ impl GameloopTimeState
             cap_frametime : Duration::new(0, 0),
             frame_start_jiffies : Duration::new(0, 0),
             last_display : Instant::now(),
-            #[cfg(feature = "imgui_base")]
+            #[cfg(feature = "imgui")]
             last_imgui : Instant::now(),
             lag_this_frame : 0,
             lag_this_second : 0,
@@ -896,13 +925,9 @@ impl GameloopTimeState
             let remaining = framerate_wait_target.saturating_sub(self.jiffy_queue);
 
             // Sleep as long as we reliably can, then busy wait
-            if remaining >= Duration::from_millis(BUSY_WAIT_MARGIN)
+            if remaining >= Duration::from_millis(BUSY_WAIT_MARGIN) || low_power
             {
-                std::thread::sleep(remaining.saturating_sub(Duration::from_millis(BUSY_WAIT_MARGIN)));
-            }
-            else if low_power
-            {
-                std::thread::sleep(remaining);
+                std::thread::sleep(Duration::from_millis(1));
             }
         }
     }
@@ -916,13 +941,9 @@ impl GameloopTimeState
             let remaining = framerate_wait_target.saturating_sub(current_display - self.last_display);
 
             // Sleep as long as we reliably can, then busy wait
-            if remaining >= Duration::from_millis(BUSY_WAIT_MARGIN)
+            if remaining >= Duration::from_millis(BUSY_WAIT_MARGIN) || low_power
             {
-                std::thread::sleep(remaining.saturating_sub(Duration::from_millis(BUSY_WAIT_MARGIN)));
-            }
-            else if low_power
-            {
-                std::thread::sleep(remaining);
+                std::thread::sleep(Duration::from_millis(1));
             }
 
             current_display = Instant::now();
@@ -932,6 +953,7 @@ impl GameloopTimeState
     }
 }
 
+/// Contains timing statistics such as FPS and milliseconds taken for each part of the gameloop.
 #[derive(Clone)]
 pub struct GameStats
 {
@@ -1002,11 +1024,13 @@ pub struct Gameloop
 
 impl Gameloop
 {
+    /// Creates the main game context. Will panic if startup fails for any reason.
     pub fn new(gameinfo : GameInfo) -> Gameloop
     {
         Gameloop::try_new(gameinfo).unwrap_or_else(|error| panic!("Startup failed: {}", error))
     }
 
+    /// Creates the main game context. Will return an error instead of panicking if startup fails for any reason.
     pub fn try_new(gameinfo : GameInfo) -> Result<Gameloop, InitError>
     {
         let mut gameinfo = gameinfo.clone();
@@ -1051,17 +1075,17 @@ impl Gameloop
         info!("Starting SDL version {}", version);
         
         // Subsystems init
-        let sdl_context = try_or_else!(sdl2::init(),
-            |error| Err(InitError::Sdl2InitError(Sdl2Subsystem::Main, error)));
-        let event_pump = try_or_else!(sdl_context.event_pump(),
-            |error| Err(InitError::Sdl2InitError(Sdl2Subsystem::Event, error)));
+        let sdl_context = sdl2::init().map_err(
+            |error| InitError::Sdl2InitError(Sdl2Subsystem::Main, error))?;
+        let event_pump = sdl_context.event_pump().map_err(
+            |error| InitError::Sdl2InitError(Sdl2Subsystem::Event, error))?;
         
         #[allow(unused_mut)] // mut used only by imgui feature
-        let mut control = Box::new(try_or_else!(GameControl::new(&gameinfo, GameControlOptions
+        let mut control = Box::new(GameControl::new(&gameinfo, GameControlOptions
         {
             console_history,
             sdl_context : Some(&sdl_context)
-        }), |error| Err(error)));
+        }).map_err(|error| error)?);
         
         control.renderer_mut().set_base_size(gameinfo.base_width as f32, gameinfo.base_height as f32);
         control.renderer_mut().set_window_title(&gameinfo.friendly_name);
@@ -1113,7 +1137,8 @@ impl Gameloop
         }
 
         self.time.think_timer.begin_frame_timing();
-        
+
+        #[cfg(feature = "audio")]
         self.control.sync_audio_store();
 
         self.time.lag_this_frame = 0;
@@ -1140,7 +1165,7 @@ impl Gameloop
             self.control.low_energy_timer += 1;
 
             // Prevent overflow if CPU cannot process all the frames
-            if (self.time.think_timer.jiffies_this_frame()) > self.time.target_frametime * MAX_LOGIC_FRAME_OVERRUN
+            if (self.time.think_timer.jiffies_this_frame()) > self.time.target_frametime
             {
                 while self.time.jiffy_queue > self.time.target_frametime
                 {
@@ -1170,7 +1195,7 @@ impl Gameloop
     {
         self.time.draw_timer.begin_frame_timing();
 
-        #[cfg(feature = "imgui_base")]
+        #[cfg(feature = "imgui")]
         {
             if !self.imgui_draw()
             {
@@ -1178,7 +1203,7 @@ impl Gameloop
             }
         }
 
-        #[cfg(not(feature = "imgui_base"))]
+        #[cfg(not(feature = "imgui"))]
         self.scene_draw();
 
         if self.time.thought_yet && self.should_draw()
@@ -1208,7 +1233,7 @@ impl Gameloop
         }
     }
 
-    #[cfg(feature = "imgui_base")]
+    #[cfg(feature = "imgui")]
     fn imgui_draw(&mut self) -> bool
     {
         let mut imgui_handler : Option<ImGuiHandler> = None;
@@ -1311,7 +1336,7 @@ impl Gameloop
     {
         self.control.low_energy_timer = 0;
 
-        #[cfg(feature = "imgui_base")]
+        #[cfg(feature = "imgui")]
         {
             if let Some(imgui_handler) = &mut self.control.imgui_handler
             {
