@@ -1,7 +1,6 @@
 //! Rendering functionality
 
-#[cfg(feature = "opengl")] pub mod opengl;
-#[cfg(feature = "vulkan")] pub mod vulkan;
+mod opengl;
 
 use cgmath::{Matrix3, SquareMatrix};
 
@@ -10,17 +9,14 @@ use std::rc::Rc;
 use std::cell::RefCell;
 
 use serde::{Deserialize, Serialize};
-use itertools::Itertools;
 
-use keeshond_datapack::{DataError, DataId, DataMultistore, DataObject, DataStore, PreparedStoreError, TrustPolicy};
-use keeshond_datapack::source::Source;
-
-#[cfg(feature = "imgui_base")] use imgui;
+#[cfg(feature = "imgui")] use imgui;
 
+use crate::datapack::{DataError, DataId, DataMultistore, DataObject, DataStore, PreparedStoreError, TrustPolicy};
+use crate::datapack::source::Source;
+use crate::renderer::opengl::GlDrawControl;
 use crate::scene::BaseScene;
-use crate::util::{BakedRect, NamedItemStore, Rect};
-#[cfg(feature = "opengl")] use crate::renderer::opengl::GlDrawControl;
-#[cfg(feature = "vulkan")] use crate::renderer::vulkan::VulkanDrawControl;
+use crate::util::{NamedItemStore, Rect};
 
 #[cfg(not(windows))] use std::io::Cursor;
 #[cfg(not(windows))] use sdl2::surface::Surface;
@@ -37,11 +33,11 @@ const MAX_MULTIPLIER_ADJUST : f32 = 0.9;
 pub(crate) fn png_to_raw<R : Read>(png_reader : R) -> Result<(Vec<u8>, u32, u32), RendererError>
 {
     let decoder = png::Decoder::new(png_reader);
-    let (info, mut reader) = try_or_else!(decoder.read_info(),
-        |error| Err(RendererError::LoadResourceFailed(format!("Failed to read PNG: {}", error))));
+    let (info, mut reader) = decoder.read_info().map_err(
+        |error| RendererError::LoadResourceFailed(format!("Failed to read PNG: {}", error)))?;
     let mut image_raw = vec![0; info.buffer_size()];
-    try_or_else!(reader.next_frame(&mut image_raw),
-        |error| Err(RendererError::LoadResourceFailed(format!("Failed to decode PNG: {}", error))));
+    reader.next_frame(&mut image_raw).map_err(
+        |error| RendererError::LoadResourceFailed(format!("Failed to decode PNG: {}", error)))?;
 
     Ok((image_raw, info.width, info.height))
 }
@@ -349,7 +345,7 @@ pub enum RendererError
     RendererInitFailed(String),
     #[fail(display = "Resource load failed: {}", _0)]
     LoadResourceFailed(String),
-    #[cfg(feature = "imgui_base")]
+    #[cfg(feature = "imgui")]
     #[fail(display = "Imgui renderer failed: {}", _0)]
     ImguiRendererFailed(String)
 }
@@ -441,21 +437,6 @@ pub struct DrawSpriteRawInfo
     pub alpha : f32
 }
 
-/// (ADVANCED) Information on how to create a new [Renderer]
-pub struct RendererNewInfo<'a>
-{
-    pub width : u32,
-    pub height : u32,
-    pub default_window_scale : WindowScaleMode,
-    pub texture_filtering : bool,
-    pub fullscreen : bool,
-    pub kiosk_mode : bool,
-    pub viewport_mode : ViewportMode,
-    pub backends : String,
-    pub video_subsystem : sdl2::VideoSubsystem,
-    pub resources : &'a mut DataMultistore
-}
-
 #[derive(Debug, Clone, Copy, Eq, PartialEq)]
 pub struct Color
 {
@@ -914,8 +895,8 @@ impl DataObject for Sheet
     {
         if self.image_file_dirty
         {
-            let writer = try_or_else!(source.write_file(package_name, pathname),
-                |error| Err(DataError::PackageSourceError(error)));
+            let writer = source.write_file(package_name, pathname).map_err(
+                |error| DataError::PackageSourceError(error))?;
 
             let mut encoder = png::Encoder::new(writer, self.width, self.height);
             encoder.set_depth(png::BitDepth::Eight);
@@ -972,11 +953,11 @@ impl DataObject for Sheet
         let mut new_path = String::from(pathname);
         new_path.push_str(".json");
 
-        let mut writer = try_or_else!(source.write_file(package_name, &new_path),
-                |error| Err(DataError::PackageSourceError(error)));
+        let mut writer = source.write_file(package_name, &new_path).map_err(
+                |error| DataError::PackageSourceError(error))?;
 
-        let json_string = try_or_else!(serde_json::to_string(&self.metadata),
-            |error| Err(DataError::BadData(format!("{}/{}: {}", package_name, pathname, error))));
+        let json_string = serde_json::to_string(&self.metadata).map_err(
+            |error| DataError::BadData(format!("{}/{}: {}", package_name, pathname, error)))?;
 
         match writer.write_all(json_string.as_bytes())
         {
@@ -1065,107 +1046,59 @@ impl DataObject for Shader
     }
 }
 
-trait DrawBackendEntryBase
-{
-    fn name(&self) -> &str;
-    fn try_init(&self, renderer_new_info : &mut RendererNewInfo) -> DrawControlResult;
-}
-
-#[allow(dead_code)]
-struct DrawBackendEntry<T : DrawControl>
-{
-    backend_name : String,
-    backend_type : std::marker::PhantomData<T>
-}
-
-impl<T : DrawControl> DrawBackendEntry<T>
-{
-    #[allow(dead_code)]
-    fn new(name : &str) -> DrawBackendEntry<T>
-    {
-        DrawBackendEntry::<T>
-        {
-            backend_name : name.to_string(),
-            backend_type : std::marker::PhantomData
-        }
-    }
-}
-
-impl<T : DrawControl> DrawBackendEntryBase for DrawBackendEntry<T>
-{
-    fn name(&self) -> &str
-    {
-        &self.backend_name
-    }
-    
-    fn try_init(&self, renderer_new_info : &mut RendererNewInfo) -> DrawControlResult
-    {
-        T::new(renderer_new_info)
-    }
-}
-
-pub fn available_backends() -> String
-{
-    let mut backends = String::new();
-
-    #[cfg(feature = "vulkan")]
-    backends.push_str("vulkan ");
-    #[cfg(feature = "opengl")]
-    backends.push_str("opengl ");
-
-    backends.pop();
-
-    backends
-}
-
-fn backend_entries_from_string(backends : &String) -> Vec<Box<dyn DrawBackendEntryBase>>
-{
-    #[allow(unused_mut)]
-        let mut backend_list : Vec<Box<dyn DrawBackendEntryBase>> = Vec::new();
-    let backend_names = backends.split(' ').map(|name| name.to_string()).unique();
-
-    for name in backend_names
-    {
-        match name.as_str()
-        {
-            #[cfg(feature = "vulkan")]
-            "vulkan" => backend_list.push(Box::new(DrawBackendEntry::<VulkanDrawControl>::new(&name))),
-            #[cfg(feature = "opengl")]
-            "opengl" => backend_list.push(Box::new(DrawBackendEntry::<GlDrawControl>::new(&name))),
-            _ => warn!("Unknown backend \"{}\"", name)
-        }
-    }
-
-    backend_list
-}
-
 pub type DrawControlResult = Result<(Box<dyn DrawControl>, Rc<RefCell<sdl2::video::Window>>), RendererError>;
 
+pub(crate) mod draw_control_private
+{
+    use crate::datapack::{DataMultistore, DataStore, PreparedStoreError};
+    use crate::renderer::{DrawControlResult, Shader, Sheet, ViewportMode, WindowScaleMode};
+    #[cfg(feature = "imgui")] use crate::renderer::RendererError;
+
+    #[doc(hidden)]
+    pub struct RendererNewInfo<'a>
+    {
+        pub width : u32,
+        pub height : u32,
+        pub default_window_scale : WindowScaleMode,
+        pub texture_filtering : bool,
+        pub fullscreen : bool,
+        pub kiosk_mode : bool,
+        pub viewport_mode : ViewportMode,
+        pub video_subsystem : sdl2::VideoSubsystem,
+        pub resources : &'a mut DataMultistore
+    }
+
+    #[doc(hidden)]
+    pub trait DrawControlBackend
+    {
+        fn new(renderer_new_info : &mut RendererNewInfo) -> DrawControlResult where Self : Sized;
+        fn sync_sheet_store(&mut self, sheet_store : &mut DataStore<Sheet>) -> Result<(), PreparedStoreError>;
+        fn sync_shader_store(&mut self, shader_store : &mut DataStore<Shader>) -> Result<(), PreparedStoreError>;
+        #[cfg(feature = "imgui")]
+        fn draw_imgui(&mut self, ui : imgui::Ui);
+        fn recalculate_viewport(&mut self, base_width : f32, base_height : f32);
+        fn start_drawing(&mut self);
+        fn end_drawing(&mut self);
+        fn present(&mut self);
+        #[cfg(feature = "imgui")]
+        fn init_imgui_renderer(&mut self, imgui : &mut imgui::Context) -> Result<(), RendererError>;
+        #[cfg(feature = "imgui")]
+        fn update_imgui_renderer(&mut self, imgui : &mut imgui::Context);
+    }
+}
+
+use draw_control_private::{DrawControlBackend, RendererNewInfo};
+
 /// Public interface to controlling rendering in your game
-pub trait DrawControl
+pub trait DrawControl : DrawControlBackend
 {
-    fn new(renderer_new_info : &mut RendererNewInfo) -> DrawControlResult where Self : Sized;
-    fn base_rect(&self) -> BakedRect<f32>;
     fn screen_transform(&self) -> &DrawTransform;
-    fn sync_sheet_store(&mut self, sheet_store : &mut DataStore<Sheet>) -> Result<(), PreparedStoreError>;
-    fn sync_shader_store(&mut self, shader_store : &mut DataStore<Shader>) -> Result<(), PreparedStoreError>;
     fn draw_sprite(&mut self, draw_info : &DrawSpriteInfo);
     fn draw_sprite_raw(&mut self, draw_info : &DrawSpriteRawInfo);
     fn draw_background_color(&mut self, draw_info : &DrawBackgroundColorInfo);
-    #[cfg(feature = "imgui_base")]
-    fn draw_imgui(&mut self, ui : imgui::Ui);
-    fn recalculate_viewport(&mut self, base_width : f32, base_height : f32);
-    fn start_drawing(&mut self);
-    fn end_drawing(&mut self);
-    fn present(&mut self);
-    #[cfg(feature = "imgui_base")]
-    fn init_imgui_renderer(&mut self, imgui : &mut imgui::Context) -> Result<(), RendererError>;
-    #[cfg(feature = "imgui_base")]
-    fn update_imgui_renderer(&mut self, imgui : &mut imgui::Context);
 }
 
-/// (ADVANCED) Null rendering backend, for automatic tests that don't need sound
-pub struct NullDrawControl
+struct NullDrawControl
 {
     transform : DrawTransform
 }
@@ -1181,7 +1114,7 @@ impl NullDrawControl
     }
 }
 
-impl DrawControl for NullDrawControl
+impl DrawControlBackend for NullDrawControl
 {
     fn new(renderer_new_info : &mut RendererNewInfo) -> DrawControlResult where Self : Sized
     {
@@ -1190,8 +1123,8 @@ impl DrawControl for NullDrawControl
         
         window_builder.hidden().position_centered().resizable().allow_highdpi();
         
-        let window_raw = try_or_else!(window_builder.build(),
-            |error| Err(RendererError::RendererInitFailed(format!("Could not create game window: {}", error))));
+        let window_raw = window_builder.build().map_err(
+            |error| RendererError::RendererInitFailed(format!("Could not create game window: {}", error)))?;
         let window = Rc::new(RefCell::new(window_raw));
         
         Ok((Box::new(NullDrawControl
@@ -1199,16 +1132,6 @@ impl DrawControl for NullDrawControl
             transform : DrawTransform::identity()
         }), window))
     }
-
-    fn base_rect(&self) -> BakedRect<f32>
-    {
-        BakedRect::default()
-    }
-
-    fn screen_transform(&self) -> &DrawTransform
-    {
-        &self.transform
-    }
     
     #[allow(unused_variables)]
     fn sync_sheet_store(&mut self, sheet_store : &mut DataStore<Sheet>) -> Result<(), PreparedStoreError>
@@ -1222,13 +1145,7 @@ impl DrawControl for NullDrawControl
         Ok(())
     }
 
-    #[allow(unused_variables)]
-    fn draw_sprite(&mut self, draw_info : &DrawSpriteInfo) {}
-    #[allow(unused_variables)]
-    fn draw_sprite_raw(&mut self, draw_info : &DrawSpriteRawInfo) {}
-    #[allow(unused_variables)]
-    fn draw_background_color(&mut self, draw_info : &DrawBackgroundColorInfo) {}
-    #[cfg(feature = "imgui_base")]
+    #[cfg(feature = "imgui")]
     #[allow(unused_variables)]
     fn draw_imgui(&mut self, ui : imgui::Ui) {}
     #[allow(unused_variables)]
@@ -1239,17 +1156,31 @@ impl DrawControl for NullDrawControl
     fn end_drawing(&mut self) {}
     #[allow(unused_variables)]
     fn present(&mut self) {}
-    #[cfg(feature = "imgui_base")]
+    #[cfg(feature = "imgui")]
     #[allow(unused_variables)]
     fn init_imgui_renderer(&mut self, imgui : &mut imgui::Context) -> Result<(), RendererError>
     {
         Ok(())
     }
-    #[cfg(feature = "imgui_base")]
+    #[cfg(feature = "imgui")]
     #[allow(unused_variables)]
     fn update_imgui_renderer(&mut self, imgui : &mut imgui::Context) {}
 }
 
+impl DrawControl for NullDrawControl
+{
+    fn screen_transform(&self) -> &DrawTransform
+    {
+        &self.transform
+    }
+    #[allow(unused_variables)]
+    fn draw_sprite(&mut self, draw_info : &DrawSpriteInfo) {}
+    #[allow(unused_variables)]
+    fn draw_sprite_raw(&mut self, draw_info : &DrawSpriteRawInfo) {}
+    #[allow(unused_variables)]
+    fn draw_background_color(&mut self, draw_info : &DrawBackgroundColorInfo) {}
+}
+
 #[derive(Eq, PartialEq)]
 enum OneToOneMode
 {
@@ -1272,9 +1203,9 @@ pub struct Renderer
 
 impl Renderer
 {
-    /// (ADVANCED) Creates a new instance. This is already created by the [crate::gameloop::GameControl]
+    /// (Advanced) Creates a new instance. This is already created by the [crate::gameloop::GameControl]
     ///  so you don't need to create this yourself.
-    pub fn new(renderer_new_info : Option<RendererNewInfo>) -> Result<Renderer, RendererError>
+    pub(crate) fn new(renderer_new_info : Option<RendererNewInfo>) -> Result<Renderer, RendererError>
     {
         let mut control : Box<dyn DrawControl> = Box::new(NullDrawControl::new_headless());
         let mut window = None;
@@ -1293,60 +1224,30 @@ impl Renderer
                 winapi::um::winuser::SetProcessDpiAwarenessContext(winapi::shared::windef::DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2);
             }
 
-            let backend_list = backend_entries_from_string(&new_info.backends);
-            
-            let mut backend_error = Err(RendererError::RendererInitFailed("No renderers available".to_string()));
-            
-            for backend in backend_list
+            let (out_control, out_window) = GlDrawControl::new(&mut new_info)?;
+
+            #[cfg(not(windows))]
             {
-                info!("Trying renderer backend \"{}\"...", backend.name());
-                
-                match backend.try_init(&mut new_info)
-                {
-                    Ok((out_control, out_window)) =>
-                    {
-                        control = out_control;
-                        window = Some(out_window);
-                        video_subsystem = Some(new_info.video_subsystem);
-                        fullscreen = new_info.fullscreen;
-                        kiosk_mode = new_info.kiosk_mode;
-                        one_to_one_viewport = match new_info.viewport_mode
-                        {
-                            ViewportMode::OneToOneDpiScaled => { OneToOneMode::DpiScaled }
-                            ViewportMode::OneToOneUnscaled => { OneToOneMode::Unscaled }
-                            _ => { OneToOneMode::Off }
-                        };
-                        base_width = new_info.width as f32;
-                        base_height = new_info.height as f32;
-                        
-                        info!("Using renderer backend \"{}\"", backend.name());
-                        
-                        break;
-                    },
-                    Err(error) =>
-                    {
-                        warn!("Renderer backend \"{}\" failed: {}", backend.name(), error);
-                        backend_error = Err(error);
-                    }
-                }
+                let (mut icon_raw, icon_width, icon_height) = png_to_raw(Cursor::new(include_bytes!("../res/icon.png").to_vec()))?;
+                let surface = Surface::from_data(&mut icon_raw, icon_width, icon_height, icon_width * 4, PixelFormatEnum::ABGR8888).map_err(
+                    |error| RendererError::LoadResourceFailed(format!("Failed to create raw surface: {}", error)))?;
+
+                out_window.borrow_mut().set_icon(surface);
             }
 
-            #[allow(unused_variables)]
-            if let Some(window) = &window
+            control = out_control;
+            window = Some(out_window);
+            video_subsystem = Some(new_info.video_subsystem);
+            fullscreen = new_info.fullscreen;
+            kiosk_mode = new_info.kiosk_mode;
+            one_to_one_viewport = match new_info.viewport_mode
             {
-                #[cfg(not(windows))]
-                {
-                    let (mut icon_raw, icon_width, icon_height) = png_to_raw(Cursor::new(include_bytes!("../res/icon.png").to_vec()))?;
-                    let surface = try_or_else!(Surface::from_data(&mut icon_raw, icon_width, icon_height, icon_width * 4, PixelFormatEnum::ABGR8888),
-                        |error| Err(RendererError::LoadResourceFailed(format!("Failed to create raw surface: {}", error))));
-
-                    window.borrow_mut().set_icon(surface);
-                }
-            }
-            else
-            {
-                return backend_error;
-            }
+                ViewportMode::OneToOneDpiScaled => { OneToOneMode::DpiScaled }
+                ViewportMode::OneToOneUnscaled => { OneToOneMode::Unscaled }
+                _ => { OneToOneMode::Off }
+            };
+            base_width = new_info.width as f32;
+            base_height = new_info.height as f32;
         }
         
         let mut renderer = Renderer
@@ -1410,7 +1311,7 @@ impl Renderer
         info.scissor.unwrap_or_default()
     }
     
-    pub fn draw(&mut self, scene : &mut Box<dyn BaseScene>, resources : &mut DataMultistore, interpolation : f32)
+    pub(crate) fn draw(&mut self, scene : &mut Box<dyn BaseScene>, resources : &mut DataMultistore, interpolation : f32)
     {
         self.control().sync_sheet_store(&mut resources.store_mut())
         .expect("GPU sheet store sync failed");
@@ -1423,14 +1324,14 @@ impl Renderer
         scene.draw(resources, &mut self.control, &first_transform, interpolation);
     }
 
-    #[cfg(feature = "imgui_base")]
-    pub fn update_imgui(&mut self, imgui : &mut imgui::Context)
+    #[cfg(feature = "imgui")]
+    pub(crate) fn update_imgui(&mut self, imgui : &mut imgui::Context)
     {
         self.control().update_imgui_renderer(imgui);
     }
 
-    #[cfg(feature = "imgui_base")]
-    pub fn draw_imgui(&mut self, ui : imgui::Ui, resources : &mut DataMultistore)
+    #[cfg(feature = "imgui")]
+    pub(crate) fn draw_imgui(&mut self, ui : imgui::Ui, resources : &mut DataMultistore)
     {
         self.control().sync_sheet_store(&mut resources.store_mut())
             .expect("GPU sheet store sync failed");
@@ -1440,7 +1341,7 @@ impl Renderer
         self.control().draw_imgui(ui);
     }
     
-    pub fn present(&mut self)
+    pub(crate) fn present(&mut self)
     {
         self.control.present();
     }
@@ -1523,16 +1424,20 @@ impl Renderer
         if let Some(window) = &mut self.window
         {
             let mut window_borrow = window.borrow_mut();
+            let new_mode;
             
             if enabled || self.kiosk_mode
             {
-                try_or_else!(window_borrow.set_fullscreen(sdl2::video::FullscreenType::Desktop),
-                    |error| warn!("Could not set fullscreen mode: {}", error));
+                new_mode = sdl2::video::FullscreenType::Desktop;
             }
             else
             {
-                try_or_else!(window_borrow.set_fullscreen(sdl2::video::FullscreenType::Off),
-                    |error| warn!("Could not set fullscreen mode: {}", error));
+                new_mode = sdl2::video::FullscreenType::Off;
+            }
+
+            if let Err(error) = window_borrow.set_fullscreen(new_mode)
+            {
+                warn!("Could not set fullscreen mode: {}", error);
             }
         }
     }
@@ -1542,16 +1447,20 @@ impl Renderer
         if let Some(window) = &mut self.window
         {
             let mut window_borrow = window.borrow_mut();
+            let new_mode;
             
             if window_borrow.fullscreen_state() == sdl2::video::FullscreenType::Off || self.kiosk_mode
             {
-                try_or_else!(window_borrow.set_fullscreen(sdl2::video::FullscreenType::Desktop),
-                    |error| warn!("Could not set fullscreen mode: {}", error));
+                new_mode = sdl2::video::FullscreenType::Desktop;
             }
             else
             {
-                try_or_else!(window_borrow.set_fullscreen(sdl2::video::FullscreenType::Off),
-                    |error| warn!("Could not set fullscreen mode: {}", error));
+                new_mode = sdl2::video::FullscreenType::Off;
+            }
+
+            if let Err(error) = window_borrow.set_fullscreen(new_mode)
+            {
+                warn!("Could not set fullscreen mode: {}", error);
             }
         }
     }
