@@ -1,4 +1,4 @@
-//! (ADVANCED) OpenGL 3 rendering backend
+//! OpenGL 3 rendering backend
 
 use std::io::Cursor;
 use std::rc::Rc;
@@ -11,21 +11,23 @@ use glium::program::ProgramCreationInput;
 use glium::backend::Facade;
 use crate::thirdparty::glium_sdl2::{DisplayBuild, SDL2Facade};
 
-use keeshond_datapack::{DataId, DataStore, PreparedStore, PreparedStoreError, DataPreparer};
+use crate::datapack::{DataId, DataStore, PreparedStore, PreparedStoreError, DataPreparer};
 
-use crate::renderer::{DrawControl, DrawControlResult, DrawTransform, DrawSpriteRawInfo, DrawBackgroundColorInfo, RendererNewInfo, RendererError, Sheet, Shader, ViewportMode, DrawSpriteInfo, Color};
+use crate::renderer::{DrawControl, DrawControlResult, DrawTransform, DrawSpriteRawInfo, DrawBackgroundColorInfo, RendererError, Sheet, Shader, ViewportMode, DrawSpriteInfo, Color};
+use crate::renderer::draw_control_private::{DrawControlBackend, RendererNewInfo};
 
-#[cfg(feature = "imgui_opengl")] use imgui_glium_renderer;
-#[cfg(feature = "imgui_opengl")] use glium::uniforms::SamplerBehavior;
-use crate::util::BakedRect;
+#[cfg(feature = "imgui")] use imgui_glium_renderer;
+#[cfg(feature = "imgui")] use glium::uniforms::SamplerBehavior;
 
-const INSTANCES_MAX : usize = 8192;
-const INSTANCE_BUFFER_MAX: usize = 32;
+const INSTANCES_MAX : usize = 2048; // How many instances to store in each instance buffer.
+const INSTANCE_BUFFER_MAX: usize = 4; // How many instance buffers to have.
+const INSTANCE_RUN_COARSENESS : usize = 128; // The "resolution" of start index for new batches within a given instance buffer.
+                                             // Use to tune performance of batch breaks vs VAO memory bloat
 
 const UNBORKED_IDENTITY_TRANSFORM : [f32; 9] = [2.0, 0.0, 0.0, 0.0, -2.0, 0.0, -1.0, 1.0, 1.0];
 const UNBORKED_UPSCALE_TRANSFORM : [f32; 9] = [2.0, 0.0, 0.0, 0.0, 2.0, 0.0, -1.0, -1.0, 1.0];
 
-#[derive(Copy, Clone, PartialEq)]
+#[derive(Copy, Clone, Eq, PartialEq)]
 enum RendererOp
 {
     #[allow(dead_code)]
@@ -45,7 +47,7 @@ impl Default for RendererOp
 #[derive(Copy, Clone)]
 struct Vertex
 {
-    a_position : [f32; 2],
+    a_position : [f32; 2]
 }
 
 implement_vertex!(Vertex, a_position);
@@ -58,12 +60,11 @@ struct Instance
     a_transform_col3 : [f32; 2],
     a_transform_offset : [f32; 2],
     a_tex_coords : [f32; 4],
-    a_color : [f32; 3],
-    a_alpha : f32
+    a_rgba : [f32; 4]
 }
 
 implement_vertex!(Instance, a_transform_col1, a_transform_col2, a_transform_col3,
-    a_transform_offset, a_tex_coords, a_color, a_alpha);
+    a_transform_offset, a_tex_coords, a_rgba);
 
 const VERTEX_QUAD : [Vertex; 4] =
 [
@@ -100,7 +101,7 @@ struct GpuSheetPreparer
     #[allow(dead_code)]
     default_texture : Rc<Texture2d>,
     texture_filtering : bool,
-    #[cfg(feature = "imgui_opengl")]
+    #[cfg(feature = "imgui")]
     imgui_renderer : Rc<RefCell<Option<imgui_glium_renderer::Renderer>>>
 }
 
@@ -186,7 +187,7 @@ impl GpuSheetPreparer
         }
     }
 
-    #[cfg(feature = "imgui_opengl")]
+    #[cfg(feature = "imgui")]
     fn update_imgui_texture(&mut self, id : DataId<Sheet>, handle : &mut GpuSheetHandle)
     {
         let mut imgui_borrow = self.imgui_renderer.borrow_mut();
@@ -245,7 +246,7 @@ impl DataPreparer<Sheet, GpuSheetHandle> for GpuSheetPreparer
 
         data.texture_dirty = false;
 
-        #[cfg(feature = "imgui_opengl")]
+        #[cfg(feature = "imgui")]
         self.update_imgui_texture(id, &mut handle);
 
         handle
@@ -254,7 +255,7 @@ impl DataPreparer<Sheet, GpuSheetHandle> for GpuSheetPreparer
     #[allow(unused_variables)]
     fn unprepare(&mut self, _prepared : &mut GpuSheetHandle, id : DataId<Sheet>)
     {
-        #[cfg(feature = "imgui_opengl")]
+        #[cfg(feature = "imgui")]
         {
             let mut imgui_borrow = self.imgui_renderer.borrow_mut();
 
@@ -287,7 +288,7 @@ impl DataPreparer<Sheet, GpuSheetHandle> for GpuSheetPreparer
             *prepared = self.make_sheet_handle_from_texture(data, prepared.texture.clone());
         }
 
-        #[cfg(feature = "imgui_opengl")]
+        #[cfg(feature = "imgui")]
         self.update_imgui_texture(id, prepared);
 
         data.texture_dirty = false;
@@ -336,7 +337,7 @@ impl DataPreparer<Shader, GpuShaderHandle> for GpuShaderPreparer
     }
 }
 
-#[derive(Copy, Clone, Default)]
+#[derive(Copy, Clone, Eq, PartialEq, Default)]
 struct InstanceBufferParam
 {
     op : RendererOp,
@@ -370,12 +371,19 @@ fn upscale_instance_from_transform(transform : &DrawTransform) -> Instance
         a_transform_col3 : [ transform.mat.z.x, transform.mat.z.y ],
         a_transform_offset : [ 0.0, 0.0 ],
         a_tex_coords : [0.0, 0.0, 1.0, 1.0],
-        a_color : [1.0, 1.0, 1.0],
-        a_alpha : 1.0
+        a_rgba : [1.0, 1.0, 1.0, 1.0],
     }
 }
 
-pub struct GlDrawControl
+#[derive(Copy, Clone, Default)]
+struct BatchRun
+{
+    start : usize,
+    len : usize,
+    param : InstanceBufferParam
+}
+
+pub(crate) struct GlDrawControl
 {
     display : Rc<RefCell<SDL2Facade>>,
     window : Rc<RefCell<sdl2::video::Window>>,
@@ -389,7 +397,6 @@ pub struct GlDrawControl
     quad_buffer : VertexBuffer<Vertex>,
     index_buffer : IndexBuffer<u16>,
     instance_buffer_src : Box<[Instance ; INSTANCES_MAX]>,
-    instance_buffer_param : Box<[InstanceBufferParam ; INSTANCES_MAX]>,
     instance_buffer : Vec<VertexBuffer<Instance>>,
     instance_buffer_flip : usize,
     default_texture : Rc<Texture2d>,
@@ -401,36 +408,33 @@ pub struct GlDrawControl
     pixel_texture : Option<Texture2d>,
     pixel_upscale_texture : Option<Texture2d>,
     current_index : usize,
+    index_runs : Vec<BatchRun>,
+    current_run : BatchRun,
     last_view_transform : DrawTransform,
     gpu_sheet_store : PreparedStore<Sheet, GpuSheetHandle>,
     gpu_shader_store : PreparedStore<Shader, GpuShaderHandle>,
-    #[cfg(feature = "imgui_opengl")]
+    #[cfg(feature = "imgui")]
     imgui_renderer : Rc<RefCell<Option<imgui_glium_renderer::Renderer>>>
 }
 
 impl GlDrawControl
 {
-    fn draw_buffers(&mut self)
+    fn perform_draw_batch(&mut self)
     {
         if self.current_index <= 0
         {
             return;
         }
-        
-        let last = self.current_index;
+
+        self.index_runs.push(self.current_run);
 
         self.instance_buffer[self.instance_buffer_flip].invalidate();
         self.instance_buffer[self.instance_buffer_flip].write(&self.instance_buffer_src[..]);
         
-        let mut last_drawn : usize = 0;
-        let mut current_op = RendererOp::BackgroundColor;
-        let mut current_sheet = DataId::new(0);
-        let mut current_shader = DataId::new(0);
-        
         // I can't belive Rust is making me do this
-        let buffer_params = self.instance_buffer_param.clone();
+        let index_skip_list = self.index_runs.clone();
         
-        let mut do_draw = |first : usize, last : usize, sheet : DataId<Sheet>, shader : DataId<Shader>, op : RendererOp|
+        let mut do_draw = | first : usize, last : usize, param : &InstanceBufferParam |
         {
             if first >= last
             {
@@ -440,17 +444,17 @@ impl GlDrawControl
             if let Some(frame) = &mut self.frame
             {
                 let mut texture_ref = self.default_texture.as_ref();
-                let mut shader_ref = &self.shader_list[op as usize];
+                let mut shader_ref = &self.shader_list[param.op as usize];
                 let mut mag_filter = MagnifySamplerFilter::Linear;
                 let mut min_filter = MinifySamplerFilter::Linear;
                 
-                if let Some(gpu_sheet_info) = self.gpu_sheet_store.get(sheet)
+                if let Some(gpu_sheet_info) = self.gpu_sheet_store.get(param.sheet)
                 {
                     texture_ref = gpu_sheet_info.texture.as_ref();
                     mag_filter = gpu_sheet_info.mag_filter;
                     min_filter = gpu_sheet_info.min_filter;
                 }
-                if let Some(gpu_shader_info) = self.gpu_shader_store.get(shader)
+                if let Some(gpu_shader_info) = self.gpu_shader_store.get(param.shader)
                 {
                     shader_ref = &gpu_shader_info.program;
                 }
@@ -459,7 +463,8 @@ impl GlDrawControl
                 {
                     t_texture : Sampler::new(texture_ref)
                         .magnify_filter(mag_filter).minify_filter(min_filter)
-                        .wrap_function(SamplerWrapFunction::Repeat)
+                        .wrap_function(SamplerWrapFunction::Repeat),
+                    u_tex_size : [ texture_ref.width() as f32, texture_ref.height() as f32 ]
                 };
                 
                 match self.viewport_mode
@@ -499,46 +504,49 @@ impl GlDrawControl
                             &uniform, &params).expect("Draw operation failed");
                     }
                 }
-                
-                
             }
         };
-        
-        for i in 0..last
+
+        for run in &index_skip_list
         {
-            let param = &buffer_params[i];
-            
-            if current_op == param.op && current_sheet == param.sheet
-                && current_shader == param.shader && i < last - 1
-            {
-                continue;
-            }
-            
-            do_draw(last_drawn, i, current_sheet, current_shader, current_op);
-            
-            current_op = param.op;
-            current_sheet = param.sheet;
-            current_shader = param.shader;
-            last_drawn = i;
+            do_draw(run.start, run.start + run.len, &run.param);
         }
-        
-        do_draw(last_drawn, last, current_sheet, current_shader, current_op);
-        
+
+        self.flip_instance_buffers();
+    }
+
+    fn check_batch_break(&mut self, param : InstanceBufferParam)
+    {
+        let skip_index = (self.current_index + INSTANCE_RUN_COARSENESS) / INSTANCE_RUN_COARSENESS * INSTANCE_RUN_COARSENESS;
+
+        if skip_index >= INSTANCES_MAX
+        {
+            self.perform_draw_batch();
+        }
+
+        if self.current_index == 0
+        {
+            self.current_run.param = param;
+        }
+        else if self.current_run.param != param
+        {
+            self.index_runs.push(self.current_run);
+            self.current_index = skip_index;
+            self.current_run = BatchRun { start : self.current_index, len : 0, param };
+        }
+    }
+
+    fn flip_instance_buffers(&mut self)
+    {
+        self.instance_buffer_flip = (self.instance_buffer_flip + 1) % self.instance_buffer.len();
         self.current_index = 0;
+        self.current_run = BatchRun { start : 0, len : 0, param : InstanceBufferParam::default() };
+        self.index_runs.clear();
 
-        self.flip_instance_buffers(2);
-    }
-
-    fn flush_drawing(&mut self)
-    {
-        self.draw_buffers();
-
-        self.display.borrow().get_context().flush();
-    }
-
-    fn flip_instance_buffers(&mut self, amount : usize)
-    {
-        self.instance_buffer_flip = (self.instance_buffer_flip + amount) % self.instance_buffer.len();
+        if self.instance_buffer_flip % (self.instance_buffer.len() / 2) == 0
+        {
+            self.display.borrow().get_context().flush();
+        }
     }
     
     fn update_pixel_upscale(&mut self)
@@ -553,180 +561,18 @@ impl GlDrawControl
         let pixel_surface = self.pixel_texture.as_mut().unwrap().as_surface();
         let (width, height) = pixel_surface.get_dimensions();
         
-        self.pixel_upscale_texture = Some(try_or_else!(Texture2d::empty_with_format(
+        self.pixel_upscale_texture = Some(Texture2d::empty_with_format(
             display_ref, UncompressedFloatFormat::U8U8U8U8, glium::texture::MipmapsOption::NoMipmap,
-            width * self.pixel_scale, height * self.pixel_scale),
-            |error| panic!("Could not create upscale framebuffer: {}", error)));
+            width * self.pixel_scale, height * self.pixel_scale).expect("Could not create upscale framebuffer"));
     }
 }
 
 impl DrawControl for GlDrawControl
 {
-    fn new(renderer_new_info : &mut RendererNewInfo) -> DrawControlResult where Self : Sized
-    {
-        let video_subsystem = &renderer_new_info.video_subsystem;
-
-        let (width, height) = crate::renderer::calculate_window_size(renderer_new_info.width as f32,
-                                                                     renderer_new_info.height as f32,
-                                                                     renderer_new_info.default_window_scale,
-                                                                     video_subsystem);
-
-        let mut window_builder = video_subsystem.window("", width, height);
-        
-        window_builder.hidden().position_centered().resizable().allow_highdpi();
-
-        #[cfg(not(target_os = "emscripten"))]
-        {
-            video_subsystem.gl_attr().set_context_profile(sdl2::video::GLProfile::Core);
-            video_subsystem.gl_attr().set_context_version(3, 2);
-        }
-        #[cfg(target_os = "emscripten")]
-        {
-            video_subsystem.gl_attr().set_context_profile(sdl2::video::GLProfile::GLES);
-            video_subsystem.gl_attr().set_context_version(3, 0);
-        }
-        
-        let display_raw = try_or_else!(window_builder.build_glium(),
-            |error| Err(RendererError::RendererInitFailed(format!("Could not create game window: {}", error))));
-        
-        let mut pixel_texture = None;
-        
-        match renderer_new_info.viewport_mode
-        {
-            ViewportMode::Independent | ViewportMode::OneToOneDpiScaled | ViewportMode::OneToOneUnscaled => (),
-            ViewportMode::Pixel =>
-            {
-                pixel_texture = Some(try_or_else!(Texture2d::empty_with_format(
-                    &display_raw, UncompressedFloatFormat::U8U8U8U8, glium::texture::MipmapsOption::NoMipmap, renderer_new_info.width, renderer_new_info.height),
-                    |error| Err(RendererError::RendererInitFailed(format!("Could not create framebuffer: {}", error)))));
-            }
-        };
-        
-        let quad_buffer = try_or_else!(glium::VertexBuffer::new(&display_raw, &VERTEX_QUAD),
-            |error| Err(RendererError::RendererInitFailed(format!("Could not create vertex buffer: {}", error))));
-        let index_buffer = try_or_else!(glium::IndexBuffer::new(&display_raw,
-            glium::index::PrimitiveType::TrianglesList, &INDEX_QUAD),
-            |error| Err(RendererError::RendererInitFailed(format!("Could not create index buffer: {}", error))));
-
-        let mut instance_buffer = Vec::new();
-
-        for _ in 0..INSTANCE_BUFFER_MAX
-        {
-            instance_buffer.push(try_or_else!(glium::VertexBuffer::<Instance>::empty_dynamic(&display_raw, INSTANCES_MAX),
-                |error| return Err(RendererError::RendererInitFailed(format!("Could not create instance buffer: {}", error)))));
-        }
-        
-        let context = display_raw.get_context();
-        info!("Renderer: {} {}", context.get_opengl_vendor_string(), context.get_opengl_renderer_string());
-        info!("OpenGL version: {}", context.get_opengl_version_string());
-        if let Some(free_vram) = context.get_free_video_memory()
-        {
-            info!("VRAM: {} MiB free", free_vram / 1024 / 1024);
-        }
-        
-        let shader_background = try_or_else!(make_program(&display_raw,
-            &modify_shader_header(include_str!("shader/vert_standard_140.glsl")),
-            &modify_shader_header(include_str!("shader/frag_solidcolor_140.glsl"))),
-            |error| Err(RendererError::RendererInitFailed(format!("Could not load shader: {}", error))));
-        let shader_sprite = try_or_else!(make_program(&display_raw,
-            &modify_shader_header(include_str!("shader/vert_standard_140.glsl")),
-            &modify_shader_header(include_str!("shader/frag_sprite_140.glsl"))),
-            |error| Err(RendererError::RendererInitFailed(format!("Could not load shader: {}", error))));
-        
-        let shader_list = vec!(shader_background, shader_sprite);
-
-        let (default_texture, default_width, default_height) = crate::renderer::png_to_raw(
-            Cursor::new(include_bytes!("data/default.png").to_vec()))?;
-        
-        let image = RawImage2d::from_raw_rgba(default_texture, (default_width, default_height));
-        let default_texture = Rc::new(try_or_else!(Texture2d::new(&display_raw, image),
-            |error| Err(RendererError::RendererInitFailed(format!("Could not upload texture: {}", error)))));
-        
-        let window = display_raw.window_clone();
-        
-        let display = Rc::new(RefCell::new(display_raw));
-        #[cfg(feature = "imgui_opengl")]
-        let imgui_renderer = Rc::new(RefCell::new(None));
-        let sheet_data_preparer = Box::new(GpuSheetPreparer
-        {
-            display : display.clone(),
-            default_texture: default_texture.clone(),
-            texture_filtering : renderer_new_info.texture_filtering,
-            #[cfg(feature = "imgui_opengl")]
-            imgui_renderer: imgui_renderer.clone()
-        });
-        let shader_data_preparer = Box::new(GpuShaderPreparer
-        {
-            display : display.clone()
-        });
-        
-        let gpu_sheet_store = try_or_else!(PreparedStore::new(&mut renderer_new_info.resources.store_mut(),
-            sheet_data_preparer),
-            |error| Err(RendererError::LoadResourceFailed(format!("Failed to create GPU sheet handle store: {}", error))));
-        
-        let gpu_shader_store = try_or_else!(PreparedStore::new(&mut renderer_new_info.resources.store_mut(),
-            shader_data_preparer),
-            |error| Err(RendererError::LoadResourceFailed(format!("Failed to create GPU shader handle store: {}", error))));
-        
-        Ok((Box::new(GlDrawControl
-        {
-            display,
-            window : window.clone(),
-            viewport_offset : (0.0, 0.0),
-            viewport_mode : renderer_new_info.viewport_mode,
-            base_width : renderer_new_info.width as f32,
-            base_height : renderer_new_info.height as f32,
-            scale_width : 0.0,
-            scale_height : 0.0,
-            need_recalc_viewport : false,
-            quad_buffer,
-            index_buffer,
-            instance_buffer_src : Box::new([Default::default() ; INSTANCES_MAX]),
-            instance_buffer_param : Box::new([Default::default() ; INSTANCES_MAX]),
-            instance_buffer,
-            instance_buffer_flip : 0,
-            default_texture,
-            shader_list,
-            scissor : None,
-            frame : None,
-            pixel_scale : 1,
-            last_pixel_scale : 0,
-            pixel_texture,
-            pixel_upscale_texture : None,
-            current_index : 0,
-            last_view_transform : DrawTransform::identity(),
-            gpu_sheet_store,
-            gpu_shader_store,
-            #[cfg(feature = "imgui_opengl")]
-            imgui_renderer
-        }), window))
-    }
-
-    fn base_rect(&self) -> BakedRect<f32>
-    {
-        BakedRect
-        {
-            x1: 0.0,
-            y1: 0.0,
-            x2: self.base_width,
-            y2: self.base_height
-        }
-    }
-
     fn screen_transform(&self) -> &DrawTransform
     {
         &self.last_view_transform
     }
-    
-    fn sync_sheet_store(&mut self, sheet_store : &mut DataStore<Sheet>) -> Result<(), PreparedStoreError>
-    {
-        self.gpu_sheet_store.sync(sheet_store)
-    }
-    
-    fn sync_shader_store(&mut self, shader_store : &mut DataStore<Shader>) -> Result<(), PreparedStoreError>
-    {
-        self.gpu_shader_store.sync(shader_store)
-    }
 
     fn draw_sprite(&mut self, draw_info: &DrawSpriteInfo)
     {
@@ -789,14 +635,18 @@ impl DrawControl for GlDrawControl
             alpha: draw_info.alpha
         });
     }
-    
+
     fn draw_sprite_raw(&mut self, draw_info : &DrawSpriteRawInfo)
     {
-        if self.current_index >= INSTANCES_MAX
+        let instance_param = InstanceBufferParam
         {
-            self.flush_drawing();
-        }
-        
+            op : RendererOp::Sprite,
+            sheet : draw_info.sheet_id,
+            shader : draw_info.shader_id
+        };
+
+        self.check_batch_break(instance_param);
+
         let transform = &draw_info.transform;
         let instance = Instance
         {
@@ -805,27 +655,25 @@ impl DrawControl for GlDrawControl
             a_transform_col3 : [ transform.mat.z.x, transform.mat.z.y ],
             a_transform_offset : [ draw_info.transform_offset_x, draw_info.transform_offset_y ],
             a_tex_coords : [ draw_info.texture_x, draw_info.texture_y, draw_info.texture_w, draw_info.texture_h ],
-            a_color : [ draw_info.r, draw_info.g, draw_info.b ],
-            a_alpha : draw_info.alpha
+            a_rgba : [ draw_info.r, draw_info.g, draw_info.b, draw_info.alpha ],
         };
         self.instance_buffer_src[self.current_index] = instance;
-        self.instance_buffer_param[self.current_index] = InstanceBufferParam
-        {
-            op : RendererOp::Sprite,
-            sheet : draw_info.sheet_id,
-            shader : draw_info.shader_id
-        };
-        
+
         self.current_index += 1;
+        self.current_run.len += 1;
     }
-    
+
     fn draw_background_color(&mut self, draw_info : &DrawBackgroundColorInfo )
     {
-        if self.current_index >= INSTANCES_MAX
+        let instance_param = InstanceBufferParam
         {
-            self.flush_drawing();
-        }
-        
+            op : RendererOp::BackgroundColor,
+            sheet : DataId::new(0),
+            shader : draw_info.shader_id
+        };
+
+        self.check_batch_break(instance_param);
+
         let instance = Instance
         {
             a_transform_col1 : [ 2.0, 0.0 ],
@@ -833,24 +681,172 @@ impl DrawControl for GlDrawControl
             a_transform_col3 : [ -1.0, 1.0 ],
             a_transform_offset : [ 0.0, 0.0 ],
             a_tex_coords : [ 0.0, 0.0, 1.0, 1.0 ],
-            a_color : [ draw_info.r, draw_info.g, draw_info.b ],
-            a_alpha : draw_info.alpha
+            a_rgba : [ draw_info.r, draw_info.g, draw_info.b, draw_info.alpha ],
         };
         self.instance_buffer_src[self.current_index] = instance;
-        self.instance_buffer_param[self.current_index] = InstanceBufferParam
+
+        self.current_index += 1;
+        self.current_run.len += 1;
+    }
+}
+
+impl DrawControlBackend for GlDrawControl
+{
+    fn new(renderer_new_info : &mut RendererNewInfo) -> DrawControlResult where Self : Sized
+    {
+        let video_subsystem = &renderer_new_info.video_subsystem;
+
+        let (width, height) = crate::renderer::calculate_window_size(renderer_new_info.width as f32,
+                                                                     renderer_new_info.height as f32,
+                                                                     renderer_new_info.default_window_scale,
+                                                                     video_subsystem);
+
+        let mut window_builder = video_subsystem.window("", width, height);
+        
+        window_builder.hidden().position_centered().resizable().allow_highdpi();
+
+        #[cfg(not(target_os = "emscripten"))]
         {
-            op : RendererOp::BackgroundColor,
-            sheet : DataId::new(0),
-            shader : draw_info.shader_id
+            video_subsystem.gl_attr().set_context_profile(sdl2::video::GLProfile::Core);
+            video_subsystem.gl_attr().set_context_version(3, 2);
+        }
+        #[cfg(target_os = "emscripten")]
+        {
+            video_subsystem.gl_attr().set_context_profile(sdl2::video::GLProfile::GLES);
+            video_subsystem.gl_attr().set_context_version(3, 0);
+        }
+        
+        let display_raw = window_builder.build_glium().map_err(
+            |error| RendererError::RendererInitFailed(format!("Could not create game window: {}", error)))?;
+        
+        let mut pixel_texture = None;
+        
+        match renderer_new_info.viewport_mode
+        {
+            ViewportMode::Independent | ViewportMode::OneToOneDpiScaled | ViewportMode::OneToOneUnscaled => (),
+            ViewportMode::Pixel =>
+            {
+                pixel_texture = Some(Texture2d::empty_with_format(
+                    &display_raw, UncompressedFloatFormat::U8U8U8U8, glium::texture::MipmapsOption::NoMipmap, renderer_new_info.width, renderer_new_info.height).map_err(
+                    |error| RendererError::RendererInitFailed(format!("Could not create framebuffer: {}", error)))?);
+            }
         };
         
-        self.current_index += 1;
+        let quad_buffer = glium::VertexBuffer::new(&display_raw, &VERTEX_QUAD).map_err(
+            |error| RendererError::RendererInitFailed(format!("Could not create vertex buffer: {}", error)))?;
+        let index_buffer = glium::IndexBuffer::new(&display_raw,
+            glium::index::PrimitiveType::TrianglesList, &INDEX_QUAD).map_err(
+            |error| RendererError::RendererInitFailed(format!("Could not create index buffer: {}", error)))?;
+
+        let mut instance_buffer = Vec::new();
+
+        for _ in 0..INSTANCE_BUFFER_MAX
+        {
+            instance_buffer.push(glium::VertexBuffer::<Instance>::empty_dynamic(&display_raw, INSTANCES_MAX).map_err(
+                |error| RendererError::RendererInitFailed(format!("Could not create instance buffer: {}", error)))?);
+        }
+        
+        let context = display_raw.get_context();
+        info!("Renderer: {} {}", context.get_opengl_vendor_string(), context.get_opengl_renderer_string());
+        info!("OpenGL version: {}", context.get_opengl_version_string());
+        if let Some(free_vram) = context.get_free_video_memory()
+        {
+            info!("VRAM: {} MiB free", free_vram / 1024 / 1024);
+        }
+        
+        let shader_background = make_program(&display_raw,
+            &modify_shader_header(include_str!("shader/vert_standard_140.glsl")),
+            &modify_shader_header(include_str!("shader/frag_solidcolor_140.glsl"))).map_err(
+            |error| RendererError::RendererInitFailed(format!("Could not load shader: {}", error)))?;
+        let shader_sprite = make_program(&display_raw,
+            &modify_shader_header(include_str!("shader/vert_standard_140.glsl")),
+            &modify_shader_header(include_str!("shader/frag_sprite_140.glsl"))).map_err(
+            |error| RendererError::RendererInitFailed(format!("Could not load shader: {}", error)))?;
+        
+        let shader_list = vec!(shader_background, shader_sprite);
+
+        let (default_texture, default_width, default_height) = crate::renderer::png_to_raw(
+            Cursor::new(include_bytes!("data/default.png").to_vec()))?;
+        
+        let image = RawImage2d::from_raw_rgba(default_texture, (default_width, default_height));
+        let default_texture = Rc::new(Texture2d::new(&display_raw, image).map_err(
+            |error| RendererError::RendererInitFailed(format!("Could not upload texture: {}", error)))?);
+        
+        let window = display_raw.window_clone();
+        
+        let display = Rc::new(RefCell::new(display_raw));
+        #[cfg(feature = "imgui")]
+        let imgui_renderer = Rc::new(RefCell::new(None));
+        let sheet_data_preparer = Box::new(GpuSheetPreparer
+        {
+            display : display.clone(),
+            default_texture: default_texture.clone(),
+            texture_filtering : renderer_new_info.texture_filtering,
+            #[cfg(feature = "imgui")]
+            imgui_renderer: imgui_renderer.clone()
+        });
+        let shader_data_preparer = Box::new(GpuShaderPreparer
+        {
+            display : display.clone()
+        });
+        
+        let gpu_sheet_store = PreparedStore::new(&mut renderer_new_info.resources.store_mut(),
+            sheet_data_preparer).map_err(
+            |error| RendererError::LoadResourceFailed(format!("Failed to create GPU sheet handle store: {}", error)))?;
+        
+        let gpu_shader_store = PreparedStore::new(&mut renderer_new_info.resources.store_mut(),
+            shader_data_preparer).map_err(
+            |error| RendererError::LoadResourceFailed(format!("Failed to create GPU shader handle store: {}", error)))?;
+        
+        Ok((Box::new(GlDrawControl
+        {
+            display,
+            window : window.clone(),
+            viewport_offset : (0.0, 0.0),
+            viewport_mode : renderer_new_info.viewport_mode,
+            base_width : renderer_new_info.width as f32,
+            base_height : renderer_new_info.height as f32,
+            scale_width : 0.0,
+            scale_height : 0.0,
+            need_recalc_viewport : false,
+            quad_buffer,
+            index_buffer,
+            instance_buffer_src : Box::new([Default::default() ; INSTANCES_MAX]),
+            instance_buffer,
+            instance_buffer_flip : 0,
+            default_texture,
+            shader_list,
+            scissor : None,
+            frame : None,
+            pixel_scale : 1,
+            last_pixel_scale : 0,
+            pixel_texture,
+            pixel_upscale_texture : None,
+            current_index : 0,
+            index_runs : Vec::new(),
+            current_run : BatchRun { start : 0, len : 0, param : InstanceBufferParam::default() },
+            last_view_transform : DrawTransform::identity(),
+            gpu_sheet_store,
+            gpu_shader_store,
+            #[cfg(feature = "imgui")]
+            imgui_renderer,
+        }), window))
     }
     
-    #[cfg(feature = "imgui_base")]
+    fn sync_sheet_store(&mut self, sheet_store : &mut DataStore<Sheet>) -> Result<(), PreparedStoreError>
+    {
+        self.gpu_sheet_store.sync(sheet_store)
+    }
+    
+    fn sync_shader_store(&mut self, shader_store : &mut DataStore<Shader>) -> Result<(), PreparedStoreError>
+    {
+        self.gpu_shader_store.sync(shader_store)
+    }
+    
+    #[cfg(feature = "imgui")]
     fn draw_imgui(&mut self, ui : imgui::Ui)
     {
-        self.flush_drawing();
+        self.perform_draw_batch();
         
         if let Some(frame) = &mut self.frame
         {
@@ -939,12 +935,14 @@ impl DrawControl for GlDrawControl
         }
         
         self.last_view_transform = view_transform.clone();
-        self.flip_instance_buffers(1);
+        self.flip_instance_buffers();
     }
 
     fn end_drawing(&mut self)
     {
-        self.flush_drawing();
+        self.perform_draw_batch();
+
+        self.display.borrow().get_context().flush();
     }
     
     fn present(&mut self)
@@ -1017,7 +1015,7 @@ impl DrawControl for GlDrawControl
         }
     }
     
-    #[cfg(feature = "imgui_base")]
+    #[cfg(feature = "imgui")]
     fn init_imgui_renderer(&mut self, imgui : &mut imgui::Context) -> Result<(), RendererError>
     {
         let mut imgui_borrow = self.imgui_renderer.borrow_mut();
@@ -1027,9 +1025,9 @@ impl DrawControl for GlDrawControl
             return Err(RendererError::ImguiRendererFailed("Already initialized".to_string()));
         }
         
-        let mut imgui_renderer = try_or_else!(imgui_glium_renderer::Renderer::init(
-            imgui, &*self.display.borrow()),
-            |error| Err(RendererError::ImguiRendererFailed(format!("Could not initialize imgui renderer: {}", error))));
+        let mut imgui_renderer = imgui_glium_renderer::Renderer::init(
+            imgui, &*self.display.borrow()).map_err(
+            |error| RendererError::ImguiRendererFailed(format!("Could not initialize imgui renderer: {}", error)))?;
 
         imgui_renderer.textures().replace(imgui::TextureId::from(0), imgui_glium_renderer::Texture
         {
@@ -1048,7 +1046,7 @@ impl DrawControl for GlDrawControl
         Ok(())
     }
 
-    #[cfg(feature = "imgui_base")]
+    #[cfg(feature = "imgui")]
     fn update_imgui_renderer(&mut self, imgui : &mut imgui::Context)
     {
         let mut imgui_borrow = self.imgui_renderer.borrow_mut();
