@@ -1,8 +1,8 @@
+use crate::datapack::{DataId, DataMultistore};
 use crate::gameloop::{GameInfo, GameControl, GameControlOptions};
 use crate::renderer::{DrawTransform, DrawBackgroundColorInfo};
-use crate::scene::{BaseScene, SimpleScene, NullSceneType, Entity, Component, ThinkerSystem, DrawerSystem, ThinkerArgs, DrawerArgs};
+use crate::scene::{BaseScene, SimpleScene, NullSceneType, Entity, Component, ThinkerSystem, DrawerSystem, ThinkerArgs, DrawerArgs, FreezeFlags};
 use rand::seq::SliceRandom;
-use keeshond_datapack::{DataId, DataMultistore};
 
 #[derive(Debug)]
 struct FooComponent { runs : u32 }
@@ -30,6 +30,19 @@ impl BarComponent
 
 impl Component for BarComponent {}
 
+#[derive(Debug)]
+struct BazComponent { runs : u32 }
+
+impl BazComponent
+{
+    fn new() -> BazComponent
+    {
+        BazComponent { runs : 0 }
+    }
+}
+
+impl Component for BazComponent {}
+
 #[derive(Debug)]
 struct FooComponentOrdered { runs : u32 }
 
@@ -118,6 +131,42 @@ impl ThinkerSystem<NullSceneType> for FooBarSystem
     }
 }
 
+#[derive(Debug)]
+struct FooBarFreezeSystem {}
+
+impl FooBarFreezeSystem
+{
+    fn new() -> FooBarFreezeSystem
+    {
+        FooBarFreezeSystem {}
+    }
+}
+
+impl ThinkerSystem<NullSceneType> for FooBarFreezeSystem
+{
+    fn think(&mut self, args : ThinkerArgs<NullSceneType>)
+    {
+        let mut foo_store = args.components.store_mut::<FooComponent>();
+        let mut bar_store = args.components.store_mut::<BarComponent>();
+        let mut baz_store = args.components.store_mut::<BazComponent>();
+
+        for (_, foo) in foo_store.iter_mut()
+        {
+            foo.runs += 1;
+        }
+
+        for (_, bar) in bar_store.iter_all_mut()
+        {
+            bar.runs += 1;
+        }
+
+        for (_, baz) in baz_store.iter_filtered_mut(FreezeFlags::REASON_HIDDEN)
+        {
+            baz.runs += 1;
+        }
+    }
+}
+
 #[derive(Debug)]
 struct FooBarSystemOrdered {}
 
@@ -311,7 +360,7 @@ fn scene_null_entity()
 fn scene_basic_system()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
+    let mut gameloop = GameControl::new_headless(&GameInfo::default()).unwrap();
     
     let entity = scene.add_entity();
     assert_eq!(scene.add_component(&entity, FooComponent::new()), true);
@@ -347,7 +396,7 @@ fn scene_basic_system()
 fn scene_multi_component_system()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
+    let mut gameloop = GameControl::new_headless(&GameInfo::default()).unwrap();
     
     let entity1 = scene.add_entity();
     let entity2 = scene.add_entity();
@@ -391,7 +440,7 @@ fn scene_multi_component_system()
     let store = scene.component_control().store::<BarComponent>();
     assert!(store.get_entity(&entity1).is_none());
     assert!(store.get_entity(&entity2).is_none());
-    assert_eq!(store.get_entity(&entity3).unwrap().runs, 00);
+    assert_eq!(store.get_entity(&entity3).unwrap().runs, 0);
     assert_eq!(store.get_entity(&entity4).unwrap().runs, 10);
 }
 
@@ -399,7 +448,7 @@ fn scene_multi_component_system()
 fn scene_remove_entity_component()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
+    let mut gameloop = GameControl::new_headless(&GameInfo::default()).unwrap();
     
     let entity1 = scene.add_entity();
     let entity2 = scene.add_entity();
@@ -530,7 +579,7 @@ fn scene_remove_random_ordered()
     for _ in 0..100
     {
         let mut scene = SimpleScene::new();
-        let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
+        let mut gameloop = GameControl::new_headless(&GameInfo::default()).unwrap();
         let mut entities = Vec::new();
         let rng = &mut rand::thread_rng();
         
@@ -595,11 +644,120 @@ fn scene_remove_random_ordered()
     }
 }
 
+#[test]
+fn scene_freeze_component()
+{
+    let mut scene = SimpleScene::new();
+    let mut gameloop = GameControl::new_headless(&GameInfo::default()).unwrap();
+
+    scene.add_thinker_system(Box::new(FooBarFreezeSystem::new()));
+
+    let mut entities = Vec::new();
+
+    for _ in 0..1000
+    {
+        let entity = scene.add_entity();
+        scene.add_component(&entity, FooComponent::new());
+        scene.add_component(&entity, BarComponent::new());
+        scene.add_component(&entity, BazComponent::new());
+
+        entities.push(entity);
+    }
+
+    scene.think(&mut gameloop);
+
+    {
+        let mut foo_store = scene.component_control().store_mut::<FooComponent>();
+        let bar_store = scene.component_control().store::<BarComponent>();
+        let baz_store = scene.component_control().store::<BazComponent>();
+
+        for entity in &entities
+        {
+            assert_eq!(foo_store.get_entity(entity).unwrap().runs, 1);
+            assert_eq!(bar_store.get_entity(entity).unwrap().runs, 1);
+            assert_eq!(baz_store.get_entity(entity).unwrap().runs, 1);
+
+            if entity.id() % 4 > 0
+            {
+                foo_store.set_frozen(entity, FreezeFlags::REASON_HIDDEN, true);
+            }
+        }
+    }
+
+    scene.think(&mut gameloop);
+
+    {
+        let mut foo_store = scene.component_control().store_mut::<FooComponent>();
+        let mut bar_store = scene.component_control().store_mut::<BarComponent>();
+        let mut baz_store = scene.component_control().store_mut::<BazComponent>();
+
+        for entity in &entities
+        {
+            if entity.id() % 4 > 0
+            {
+                assert_eq!(foo_store.get_entity(entity).unwrap().runs, 1);
+            }
+            else
+            {
+                assert_eq!(foo_store.get_entity(entity).unwrap().runs, 2);
+            }
+            assert_eq!(bar_store.get_entity(entity).unwrap().runs, 2);
+            assert_eq!(baz_store.get_entity(entity).unwrap().runs, 2);
+
+            if entity.id() % 4 > 0
+            {
+                foo_store.set_frozen(entity, FreezeFlags::REASON_HIDDEN, false);
+                bar_store.set_frozen(entity, FreezeFlags::REASON_HIDDEN, true);
+            }
+            if entity.id() % 2 == 1
+            {
+                baz_store.set_frozen(entity, FreezeFlags::REASON_HIDDEN, true);
+            }
+            if entity.id() % 4 > 1
+            {
+                baz_store.set_frozen(entity, FreezeFlags::REASON_EFFECT, true);
+            }
+        }
+    }
+
+    scene.think(&mut gameloop);
+
+    {
+        let foo_store = scene.component_control().store::<FooComponent>();
+        let bar_store = scene.component_control().store::<BarComponent>();
+        let baz_store = scene.component_control().store::<BazComponent>();
+
+        for entity in &entities
+        {
+            let expected_baz = match entity.id() % 4
+            {
+                0 => {3} // none
+                1 => {2} // REASON_HIDDEN
+                2 => {3} // REASON_EFFECT
+                3 => {2} // REASON_EFFECT and REASON_HIDDEN
+                _ => unreachable!()
+            };
+
+            if entity.id() % 4 > 0
+            {
+                assert_eq!(foo_store.get_entity(entity).unwrap().runs, 2);
+            }
+            else
+            {
+                assert_eq!(foo_store.get_entity(entity).unwrap().runs, 3);
+                assert_eq!(baz_store.get_entity(entity).unwrap().runs, 3);
+            }
+            assert_eq!(bar_store.get_entity(entity).unwrap().runs, 3);
+            assert_eq!(baz_store.get_entity(entity).unwrap().runs, expected_baz);
+        }
+    }
+}
+
 #[test]
 fn scene_component_registration()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
+    let mut gameloop = GameControl::new_headless(&GameInfo::default()).unwrap();
 
     scene.add_thinker_system(Box::new(FooSystem::new()));
     scene.add_thinker_system(Box::new(BarAdderSystem::new()));
@@ -627,7 +785,7 @@ fn scene_component_registration()
 fn scene_immutable_draw()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
+    let mut gameloop = GameControl::new_headless(&GameInfo::default()).unwrap();
     
     let entity1 = scene.add_entity();
     let entity2 = scene.add_entity();
@@ -647,11 +805,11 @@ fn scene_immutable_draw()
 }
 
 #[test]
-#[should_panic(expected = "Mutable component access is not allowed while drawing.")]
+#[should_panic(expected = "ComponentStore for keeshond::tests::scene::FooComponent unavailable: Mutable access is not allowed while drawing")]
 fn scene_mutable_draw_should_panic()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
+    let mut gameloop = GameControl::new_headless(&GameInfo::default()).unwrap();
     
     let entity1 = scene.add_entity();
     let entity2 = scene.add_entity();
@@ -682,6 +840,7 @@ struct DeferredSystem
     foobar2_entity : Entity,
     ghost_entity : Entity,
     spaced_entity : Entity,
+    frozen_entity : Entity
 }
 
 impl DeferredSystem
@@ -697,6 +856,7 @@ impl DeferredSystem
             foobar2_entity : Entity::null(),
             ghost_entity : Entity::null(),
             spaced_entity : Entity::null(),
+            frozen_entity : Entity::null()
         }
     }
 }
@@ -726,6 +886,10 @@ impl ThinkerSystem<NullSceneType> for DeferredSystem
                 self.foobar2_entity = args.scene.add_entity_later();
                 args.scene.add_component_later(&self.foobar2_entity, FooComponent::new());
                 args.scene.add_component_later(&self.foobar2_entity, BarComponent { runs : 3 });
+                self.frozen_entity = args.scene.add_entity_later();
+                args.scene.add_component_later(&self.frozen_entity, FooComponent::new());
+                args.scene.add_component_later(&self.frozen_entity, BarComponent::new());
+                args.scene.set_entity_frozen_later(&self.frozen_entity, FreezeFlags::REASON_HIDDEN, true);
 
                 assert_eq!(foo_store.count(), 0);
                 assert_eq!(bar_store.count(), 0);
@@ -734,20 +898,26 @@ impl ThinkerSystem<NullSceneType> for DeferredSystem
                 assert!(bar_store.get_entity(&self.foobar_entity).is_none());
                 assert!(foo_store.get_entity(&self.foobar2_entity).is_none());
                 assert!(bar_store.get_entity(&self.foobar2_entity).is_none());
+                assert!(foo_store.freeze_flags(&self.frozen_entity).is_none());
+                assert!(bar_store.freeze_flags(&self.frozen_entity).is_none());
             }
             1 =>
             {
                 let foo_store = args.components.store_mut::<FooComponent>();
                 let bar_store = args.components.store_mut::<BarComponent>();
 
-                assert_eq!(foo_store.count(), 3);
-                assert_eq!(bar_store.count(), 2);
+                assert_eq!(foo_store.count(), 4);
+                assert_eq!(bar_store.count(), 3);
                 assert!(foo_store.get_entity(&self.foo_entity).is_some());
                 assert!(foo_store.get_entity(&self.foobar_entity).is_some());
                 assert!(foo_store.get_entity(&self.foobar2_entity).is_some());
                 assert!(bar_store.get_entity(&self.foo_entity).is_none());
                 assert!(bar_store.get_entity(&self.foobar_entity).is_some());
                 assert!(bar_store.get_entity(&self.foobar2_entity).is_some());
+                assert!(foo_store.freeze_flags(&self.frozen_entity).is_some());
+                assert!(bar_store.freeze_flags(&self.frozen_entity).is_some());
+                assert_eq!(foo_store.freeze_flags(&self.frozen_entity).unwrap(), FreezeFlags::REASON_HIDDEN);
+                assert_eq!(bar_store.freeze_flags(&self.frozen_entity).unwrap(), FreezeFlags::REASON_HIDDEN);
 
                 let bar = bar_store.get_entity(&self.foobar_entity).unwrap();
 
@@ -756,29 +926,35 @@ impl ThinkerSystem<NullSceneType> for DeferredSystem
 
                 args.scene.remove_component_later::<FooComponent>(&self.foobar_entity);
                 args.scene.remove_entity_later(&self.foobar2_entity);
+                args.scene.set_entity_frozen_later(&self.frozen_entity, FreezeFlags::REASON_HIDDEN, false);
+                args.scene.set_entity_frozen_later(&self.frozen_entity, FreezeFlags::REASON_EFFECT, true);
 
-                assert_eq!(foo_store.count(), 3);
-                assert_eq!(bar_store.count(), 2);
+                assert_eq!(foo_store.count(), 4);
+                assert_eq!(bar_store.count(), 3);
                 assert!(foo_store.get_entity(&self.foo_entity).is_some());
                 assert!(foo_store.get_entity(&self.foobar_entity).is_some());
                 assert!(foo_store.get_entity(&self.foobar2_entity).is_some());
                 assert!(bar_store.get_entity(&self.foo_entity).is_none());
                 assert!(bar_store.get_entity(&self.foobar_entity).is_some());
                 assert!(bar_store.get_entity(&self.foobar2_entity).is_some());
+                assert_eq!(foo_store.freeze_flags(&self.frozen_entity).unwrap(), FreezeFlags::REASON_HIDDEN);
+                assert_eq!(bar_store.freeze_flags(&self.frozen_entity).unwrap(), FreezeFlags::REASON_HIDDEN);
             }
             2 =>
             {
                 let foo_store = args.components.store_mut::<FooComponent>();
                 let bar_store = args.components.store_mut::<BarComponent>();
 
-                assert_eq!(foo_store.count(), 1);
-                assert_eq!(bar_store.count(), 1);
+                assert_eq!(foo_store.count(), 2);
+                assert_eq!(bar_store.count(), 2);
                 assert!(foo_store.get_entity(&self.foo_entity).is_some());
                 assert!(foo_store.get_entity(&self.foobar_entity).is_none());
                 assert!(foo_store.get_entity(&self.foobar2_entity).is_none());
                 assert!(bar_store.get_entity(&self.foo_entity).is_none());
                 assert!(bar_store.get_entity(&self.foobar_entity).is_some());
                 assert!(bar_store.get_entity(&self.foobar2_entity).is_none());
+                assert_eq!(foo_store.freeze_flags(&self.frozen_entity).unwrap(), FreezeFlags::REASON_EFFECT);
+                assert_eq!(bar_store.freeze_flags(&self.frozen_entity).unwrap(), FreezeFlags::REASON_EFFECT);
 
                 args.scene.add_component_later(&self.foobar_entity, FooComponent { runs : 4 });
                 self.foobar2_entity = args.scene.add_entity_later();
@@ -790,8 +966,8 @@ impl ThinkerSystem<NullSceneType> for DeferredSystem
                 let foo_store = args.components.store_mut::<FooComponent>();
                 let bar_store = args.components.store_mut::<BarComponent>();
 
-                assert_eq!(foo_store.count(), 3);
-                assert_eq!(bar_store.count(), 2);
+                assert_eq!(foo_store.count(), 4);
+                assert_eq!(bar_store.count(), 3);
                 assert!(foo_store.get_entity(&self.foo_entity).is_some());
                 assert!(foo_store.get_entity(&self.foobar_entity).is_some());
                 assert!(foo_store.get_entity(&self.foobar2_entity).is_some());
@@ -815,8 +991,8 @@ impl ThinkerSystem<NullSceneType> for DeferredSystem
                 let foo_store = args.components.store_mut::<FooComponent>();
                 let bar_store = args.components.store_mut::<BarComponent>();
 
-                assert_eq!(foo_store.count(), 3);
-                assert_eq!(bar_store.count(), 2);
+                assert_eq!(foo_store.count(), 4);
+                assert_eq!(bar_store.count(), 3);
 
                 assert!(foo_store.get_entity(&self.ghost_entity).is_none());
                 assert!(bar_store.get_entity(&self.ghost_entity).is_none());
@@ -836,8 +1012,8 @@ impl ThinkerSystem<NullSceneType> for DeferredSystem
                 let foo_store = args.components.store_mut::<FooComponent>();
                 let bar_store = args.components.store_mut::<BarComponent>();
 
-                assert_eq!(foo_store.count(), 3);
-                assert_eq!(bar_store.count(), 3);
+                assert_eq!(foo_store.count(), 4);
+                assert_eq!(bar_store.count(), 4);
                 assert!(bar_store.get_entity(&self.spaced_entity).is_some());
 
                 let bar = bar_store.get_entity(&self.spaced_entity).unwrap();
@@ -852,23 +1028,23 @@ impl ThinkerSystem<NullSceneType> for DeferredSystem
 }
 
 #[test]
-fn scene_deferred_add_remove()
+fn scene_deferred_add_remove_freeze()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
+    let mut gameloop = GameControl::new_headless(&GameInfo::default()).unwrap();
 
     scene.add_thinker_system(Box::new(DeferredSystem::new()));
 
     assert_eq!(scene.entity_count(), 0);
     scene.think(&mut gameloop);
-    assert_eq!(scene.entity_count(), 4);
-    scene.think(&mut gameloop);
-    assert_eq!(scene.entity_count(), 3);
+    assert_eq!(scene.entity_count(), 5);
     scene.think(&mut gameloop);
     assert_eq!(scene.entity_count(), 4);
     scene.think(&mut gameloop);
-    assert_eq!(scene.entity_count(), 4);
+    assert_eq!(scene.entity_count(), 5);
     scene.think(&mut gameloop);
-    assert_eq!(scene.entity_count(), 6);
+    assert_eq!(scene.entity_count(), 5);
+    scene.think(&mut gameloop);
+    assert_eq!(scene.entity_count(), 7);
     scene.think(&mut gameloop);
 }
