@@ -94,7 +94,7 @@ impl<T : Debug + Copy + Add<Output = T> + AddAssign + SubAssign + MulAssign + Pa
 
     pub fn contains(&self, x : T, y : T) -> bool
     {
-        x >= self.x && x < self.x + self.w && y >= self.y && y < self.y + self.h
+        x >= self.x && x <= self.x + self.w && y >= self.y && y <= self.y + self.h
     }
 }
 
@@ -167,14 +167,17 @@ impl<T : Debug + Copy + Sub<Output = T> + AddAssign + SubAssign + MulAssign + Pa
 
     pub fn contains(&self, x : T, y : T) -> bool
     {
-        x >= self.x1 && x < self.x2 && y >= self.y1 && y < self.y2
+        x >= self.x1 && x <= self.x2 && y >= self.y1 && y <= self.y2
+    }
+
+    pub fn touches(&self, other : &BakedRect<T>) -> bool
+    {
+        self.x1 <= other.x2 && self.x2 >= other.x1 && self.y1 <= other.y2 && self.y2 >= other.y1
     }
 
     pub fn overlaps(&self, other : &BakedRect<T>) -> bool
     {
-        let result = self.x1 < other.x2 && self.x2 >= other.x1 && self.y1 < other.y2 && self.y2 >= other.y1;
-
-        result
+        self.x1 < other.x2 && self.x2 > other.x1 && self.y1 < other.y2 && self.y2 > other.y1
     }
 }
 
